from . import modeling_layer
from . import randalo
from . import reductions
from . import truncnorm
from . import utils

from .randalo import RandALO
from .modeling_layer import HyperParameter, Regularizer, SquareRegularizer, L1Regularizer, L2Regularizer, Loss, LogisticLoss, MSELoss
from .reductions import Jacobian, gen_cvxpy_jacobian



import unittest

import numpy as np
import torch

from randalo import utils


class TestUtils(unittest.TestCase):
    def test_to_tensor(self):
        x = [1, 2, 3]
        x_np = np.array(x)
        self.assertIs(x_np.dtype, np.dtype("int64"))
        x_torch = torch.tensor(x)
        self.assertIs(x_torch.dtype, torch.int64)
        x_torch = x_torch.to(torch.float32)

        x_to_tensor = utils.to_tensor(x)
        self.assertTrue(torch.allclose(x_torch, x_to_tensor))
        self.assertIs(x_to_tensor.dtype, torch.float32)

        x_np_to_tensor = utils.to_tensor(x_np)
        self.assertTrue(torch.allclose(x_torch, x_np_to_tensor))
        self.assertIs(x_np_to_tensor.dtype, torch.float32)

        x_torch_to_tensor = utils.to_tensor(x_torch)
        self.assertTrue(torch.allclose(x_torch, x_torch_to_tensor))
        self.assertIs(x_torch_to_tensor.dtype, torch.float32)

        with self.assertRaises(ValueError):
            utils.to_tensor(1)

    def test_compute_derivatives(self):
        n = 100

        def loss_fun(y, z):
            return torch.sum((y - z) ** 4)

        rng = torch.Generator().manual_seed(0)
        y = torch.randn(n, generator=rng)
        z = torch.randn(n, generator=rng)

        y_, z_, dloss_dz, d2loss_dboth, d2loss_dz2 = utils.compute_derivatives(
            loss_fun, y, z
        )
        self.assertTrue(torch.allclose(y, y_))
        self.assertTrue(torch.allclose(z, z_))
        self.assertTrue(torch.allclose(4 * (z - y) ** 3, dloss_dz))
        self.assertTrue(torch.allclose(-12 * (y - z) ** 2, d2loss_dboth))
        self.assertTrue(torch.allclose(12 * (y - z) ** 2, d2loss_dz2))

    def test_unsqueeze_scalar_like(self):
        x = 1.5
        arr1 = torch.ones(10, dtype=torch.float32)
        arr2 = torch.ones(10, 10, dtype=torch.float64)
        arr3 = torch.ones(10, 10, 10, dtype=torch.int32)

        x1 = utils.unsqueeze_scalar_like(x, arr1)
        self.assertEqual(x1.shape, (1,))
        self.assertIs(x1.dtype, torch.float32)
        self.assertEqual(x1.item(), x)

        x2 = utils.unsqueeze_scalar_like(x, arr2)
        self.assertEqual(x2.shape, (1, 1))
        self.assertIs(x2.dtype, torch.float64)
        self.assertEqual(x2.item(), x)

        x3 = utils.unsqueeze_scalar_like(x, arr3)
        self.assertEqual(x3.shape, (1, 1, 1))
        self.assertIs(x3.dtype, torch.int32)
        self.assertNotEqual(x3.item(), x)
        self.assertEqual(x3.item(), np.floor(x))

    def test_create_mixing_matrix(self):
        m = 5
        subsets = [[0, 1, 2], [2, 3], [4]]
        M0 = torch.tensor(
            [[1 / 3, 0, 0], [1 / 3, 0, 0], [1 / 3, 1 / 2, 0], [0, 1 / 2, 0], [0, 0, 1]]
        )
        M = utils.create_mixing_matrix(m, subsets)
        self.assertTrue(torch.allclose(M0, M))

    def test_robust_y_intercept(self):
        n = 100
        rng = torch.Generator().manual_seed(0)
        x = torch.randn(n // 2, generator=rng)
        y0 = np.pi
        z = torch.randn(n // 2, generator=rng)
        y = torch.concatenate([y0 + x + z, y0 + x - z])
        x = torch.concatenate([x, x])

        y0_hat = utils.robust_y_intercept(x, y)
        self.assertAlmostEqual(y0, y0_hat, places=4)


if __name__ == "__main__":
    unittest.main()



import math

import torch

"""Implementation of truncated normal distributions in PyTorch.

This module is based on the implementation in scipy.stats.truncnorm, which is
licensed under the BSD license. The original source code can be found at
https://github.com/scipy/scipy/blob/v1.11.3/scipy/stats/_continuous_distns.py

Copyright (c) 2001-2002 Enthought, Inc. 2003-2023, SciPy Developers.
All rights reserved.
"""

# Constants for the standard normal distribution
_norm_pdf_C = math.sqrt(2 * math.pi)
_norm_pdf_logC = math.log(_norm_pdf_C)


def _norm_logpdf(x: torch.Tensor) -> torch.Tensor:
    """Log of the probability density function of the standard normal distribution."""
    return -(x**2) / 2.0 - _norm_pdf_logC


def _log_diff(log_a: torch.Tensor, log_b: torch.Tensor) -> torch.Tensor:
    """Compute log(a - b) when log_a > log_b and both are log-space values"""
    return log_a + torch.log1p(-torch.exp(log_b - log_a))


def _log_gauss_mass(alpha: torch.Tensor, beta: torch.Tensor) -> torch.Tensor:
    """Log of Gaussian probability mass within an interval"""

    alpha, beta = torch.broadcast_tensors(alpha, beta)

    # Calculations in right tail are inaccurate, so we'll exploit the
    # symmetry and work only in the left tail
    case_left = beta <= 0
    case_right = alpha > 0
    case_central = ~(case_left | case_right)

    def mass_case_left(a, b):
        return _log_diff(torch.special.log_ndtr(b), torch.special.log_ndtr(a))

    def mass_case_right(a, b):
        return mass_case_left(-b, -a)

    def mass_case_central(a, b):
        return torch.log1p(-torch.special.ndtr(a) - torch.special.ndtr(-b))

    # Initialize tensor for output
    out = torch.full_like(alpha, float("nan"))
    out[case_left] = mass_case_left(alpha[case_left], beta[case_left]).real
    out[case_right] = mass_case_right(alpha[case_right], beta[case_right]).real
    out[case_central] = mass_case_central(alpha[case_central], beta[case_central])

    return out


def _truncnorm_log_pdf(x, alpha, beta):
    """Log of the probability density function of the truncated normal distribution."""
    return _norm_logpdf(x) - _log_gauss_mass(alpha, beta)


def _truncnorm_pdf(x, alpha, beta):
    """Probability density function of the truncated normal distribution."""
    return torch.exp(_truncnorm_log_pdf(x, alpha, beta))


def truncnorm_mean(
    mu: torch.Tensor, sigma: torch.Tensor, a: torch.Tensor, b: torch.Tensor
) -> torch.Tensor:
    """Calculates the mean of a truncated normal distribution given its parameters.

    Parameters
    ----------
    mu : torch.Tensor
        The means of the normal distributions
    sigma : torch.Tensor
        The standard deviations of the normal distributions
    a : torch.Tensor
        The lower bounds of the truncation interval
    b : torch.Tensor
        The upper bounds of the truncation interval

    Returns
    -------
    torch.Tensor
        The means of the truncated normal distributions
    """

    mu, sigma, a, b = torch.broadcast_tensors(mu, sigma, a, b)

    alpha = (a - mu) / sigma
    beta = (b - mu) / sigma

    nan = torch.isnan(alpha) | torch.isnan(beta)
    notnan = ~nan
    alpha = alpha[notnan]
    beta = beta[notnan]
    alpha_beta = torch.stack([alpha, beta], dim=0)
    pA, pB = _truncnorm_pdf(alpha_beta, alpha[None, :], beta[None, :])

    out = torch.full_like(mu, float("nan"))
    out[nan] = torch.clamp(mu[nan], min=a[nan], max=b[nan])
    out[notnan] = mu[notnan] - sigma[notnan] * (pB - pA)

    return out



from typing import Callable, NamedTuple

import numpy as np
import sklearn.linear_model
import torch
from torch import autograd


def to_tensor(
    array: np.ndarray | torch.Tensor | list[float],
    dtype: torch.dtype = torch.float32,
    device: torch.device = None,
) -> torch.Tensor:
    """Convert a numpy array or torch tensor to a torch tensor.

    Parameters
    ----------
    array : np.ndarray | torch.Tensor | list[float]
        Input array or tensor.
    dtype : torch.dtype, optional
        Data type for the output tensor, by default torch.float32.
    device : torch.device, optional
        Device for the output tensor, by default None.

    Returns
    -------
    torch.Tensor
        Output tensor.
    """

    if isinstance(array, np.ndarray):
        return torch.tensor(array, dtype=dtype, device=device)
    elif isinstance(array, torch.Tensor) or isinstance(array, list):
        return torch.as_tensor(array, dtype=dtype, device=device)
    else:
        raise ValueError("Input must be a numpy array or torch tensor")


class LossDerivatives(NamedTuple):
    y: torch.Tensor
    y_hat: torch.Tensor
    y: torch.Tensor
    dloss_dy_hat: torch.Tensor
    d2loss_dboth: torch.Tensor
    d2loss_dy_hat2: torch.Tensor


def compute_derivatives(
    loss_fun: Callable[[torch.Tensor, torch.Tensor], torch.Tensor],
    y: torch.Tensor,
    y_hat: torch.Tensor,
) -> LossDerivatives:
    """Compute first and second derivatives of a loss function.

    Parameters
    ----------
    loss_fun : Callable[[torch.Tensor, torch.Tensor], torch.Tensor]
        Loss function to compute derivatives of. The function should take the
        true labels `y` and the predicted values `y_hat` as input and return
        the sum or mean reduction of the element-wise losses as a singleton
        `torch.Tensor`.
    y : torch.Tensor
        True labels.
    y_hat : torch.Tensor
        Predicted values.

    Returns
    -------
    tuple[torch.Tensor, torch.Tensor, torch.Tensor, torch.Tensor, torch.Tensor]
        The true labels,
        the predicted values,
        the first derivatives of the loss with respect to the predicted values,
        the second derivatives of the loss with respect to both the true labels
            and predicted values,
        and the second derivatives of the loss with respect to the predicted values.
    """

    # detach and clone to avoid memory leaks
    y = y.detach().clone().requires_grad_(True)
    y_hat = y_hat.detach().clone().requires_grad_(True)

    # compute first and second derivatives of loss function
    # we obtain the vector derivatives by summing and then taking the gradient
    loss = loss_fun(y, y_hat)
    # keep the graph for computing the second derivatives
    dloss_dy_hat, *_ = autograd.grad(loss, y_hat, create_graph=True)
    dloss_dy_hat_sum = dloss_dy_hat.sum()

    d2loss_dboth, d2loss_dy_hat2 = autograd.grad(
        dloss_dy_hat_sum, [y, y_hat], allow_unused=True
    )
    if d2loss_dboth is None:
        d2loss_dboth = torch.zeros_like(y)
    if d2loss_dy_hat2 is None:
        d2loss_dy_hat2 = torch.zeros_like(y_hat)

    # free memory used by autograd and return
    return LossDerivatives(
        y.detach(),
        y_hat.detach(),
        dloss_dy_hat.detach(),
        d2loss_dboth.detach(),
        d2loss_dy_hat2.detach(),
    )


def unsqueeze_scalar_like(x: float, array: torch.Tensor) -> torch.Tensor:
    """Expand a scalar to the number of dimensions of an array.

    Parameters
    ----------
    x : float
        Scalar value.
    array : torch.Tensor
        Array to expand the scalar to.

    Returns
    -------
    torch.Tensor
        Expanded scalar.
    """

    return torch.tensor(x, dtype=array.dtype, device=array.device).reshape(
        *(1,) * array.ndim
    )


def create_mixing_matrix(m: int, subsets: list[list[int]]) -> torch.Tensor:
    """Create a mixing matrix A for a given set of subsets, such that when
    M @ A is computed, the columns of M corresponding to the subsets are
    averaged.

    Parameters
    ----------
    m : int
        Number of rows of the mixing matrix.
    subsets : list[list[int]]
        List of subsets of [m] to average.

    Returns
    -------
    torch.Tensor
        Mixing matrix.
    """

    # create mixing matrix
    A = torch.zeros(m, len(subsets))
    for j, subset in enumerate(subsets):
        A[subset, j] = 1 / len(subset)

    return A


def robust_y_intercept(
    x: torch.Tensor | np.ndarray | list[float],
    y: torch.Tensor | np.ndarray | list[float],
    epsilon: float = 1.35,
) -> float:
    """Find the y-intercept of the robust linear fit.

    Parameters
    ----------
    x : torch.Tensor | np.ndarray | list[float]
        Input data.
    y : torch.Tensor | np.ndarray | list[float]
        Output data.
    epsilon : float, optional
        Huber loss parameter, by default 1.35.

    Returns
    -------
    float
        The y-intercept.
    """
    x = np.asarray(x).reshape(-1, 1)
    y = np.asarray(y).reshape(-1)

    huber = sklearn.linear_model.HuberRegressor(fit_intercept=True, epsilon=epsilon)
    huber.fit(x, y)
    return huber.intercept_



import unittest

import numpy as np
import scipy
import torch

from randalo import truncnorm


class TruncnormData(object):
    def __init__(self, mu, sigma, a=None, b=None, alpha=None, beta=None):
        self.mu = mu
        self.sigma = sigma

        match (a, b, alpha, beta):
            case (None, None, None, None):
                self.a = -np.inf
                self.b = np.inf
                self.alpha = -np.inf
                self.beta = np.inf

            case (a, b, None, None):
                self.a = a
                self.b = b
                self.alpha = (a - mu) / sigma
                self.beta = (b - mu) / sigma

            case (None, None, alpha, beta):
                self.alpha = alpha
                self.beta = beta
                self.a = mu + alpha * sigma
                self.b = mu + beta * sigma

            case (a, b, alpha, beta):
                raise ValueError("Cannot specify both a, b and alpha, beta")

    def get_torch_tensors(self, dtype=torch.float64):
        return (
            torch.tensor(x, dtype=dtype) for x in [self.mu, self.sigma, self.a, self.b]
        )


class TestTruncnorm(unittest.TestCase):
    def test_truncnorm_mean(self):
        rng = np.random.default_rng(0)

        # test basic truncnorm_mean
        td = TruncnormData(mu=1.0, sigma=1.0, a=-1.0, b=1.0)
        mean_scipy = scipy.stats.truncnorm.mean(
            td.alpha, td.beta, loc=td.mu, scale=td.sigma
        )
        mean_torch = truncnorm.truncnorm_mean(*td.get_torch_tensors()).item()
        self.assertAlmostEqual(mean_scipy, mean_torch, places=10)

        # test truncnorm_mean with no truncation
        td = TruncnormData(mu=1.0, sigma=1.0)
        mean_scipy = scipy.stats.truncnorm.mean(
            td.alpha, td.beta, loc=td.mu, scale=td.sigma
        )
        mean_torch = truncnorm.truncnorm_mean(*td.get_torch_tensors()).item()
        self.assertAlmostEqual(td.mu, mean_torch, places=10)
        self.assertAlmostEqual(mean_scipy, mean_torch, places=10)

        # test truncnorm_mean with broadcasting
        n = 5
        m = 6
        k = 7
        mu = rng.uniform(-1, 1, size=(n, m, k))
        sigma = rng.uniform(0.1, 1, size=(n, m, 1))
        a = rng.uniform(-2, -1, size=(n, 1, k))
        b = rng.uniform(1, 2, size=(1, m, 1))
        td = TruncnormData(mu=mu, sigma=sigma, a=a, b=b)
        mean_scipy = scipy.stats.truncnorm.mean(
            td.alpha, td.beta, loc=td.mu, scale=td.sigma
        )
        mean_torch = truncnorm.truncnorm_mean(*td.get_torch_tensors()).numpy()
        self.assertTrue(np.allclose(mean_scipy, mean_torch))


if __name__ == "__main__":
    unittest.main()



import unittest

import linops as lo
import numpy as np
import torch

from randalo import RandALO
from randalo import modeling_layer as ml
from randalo import truncnorm
from randalo import utils


class TestRandALO(unittest.TestCase):
    def setUp(self):
        self.n = 10
        self.rng = torch.Generator().manual_seed(0)
        self.diag = torch.rand(self.n, generator=self.rng)
        self.jac = lo.DiagonalOperator(self.diag)
        self.loss = ml.MSELoss()
        self.y = torch.rand(self.n, generator=self.rng)
        self.y_hat = torch.rand(self.n, generator=self.rng)

    def risk_fun(self, y, z):
        return torch.mean((y - z) ** 2).item()

    def test_diagonal_jac(self):
        ra = RandALO(
            loss=self.loss, jac=self.jac, y=self.y, y_hat=self.y_hat, rng=self.rng
        )
        risk = ra.evaluate(self.risk_fun, n_matvecs=3)
        self.assertEqual(ra._n_matvecs, 3)
        self.assertEqual(ra._normalized_diag_jac_estims.shape, (self.n, 3))
        self.assertTrue(
            torch.allclose(ra._normalized_diag_jac_estims, self.diag[:, None])
        )
        self.assertTrue(
            torch.allclose(ra._normalized_diag_jac_stds, torch.zeros(self.n, 1))
        )

        risk_bks = ra.evaluate_bks(self.risk_fun, n_matvecs=3)
        risk_alo = ra.evaluate_alo(self.risk_fun)
        self.assertAlmostEqual(risk, risk_bks)
        self.assertAlmostEqual(risk, risk_alo)

        self.assertTrue(
            torch.allclose(
                ra._y_tilde_exact,
                self.y_hat + self.diag / (1 - self.diag) * (self.y_hat - self.y),
            )
        )
        self.assertAlmostEqual(
            risk,
            torch.mean((self.y - self.y_hat) ** 2 / (1 - self.diag) ** 2).item(),
        )

    def test_psd_jac(self):
        X = torch.rand(self.n, self.n, generator=self.rng)
        J = X @ torch.linalg.solve(X.T @ X + torch.eye(self.n), X.T)
        diag = torch.diag(J)
        jac = lo.MatrixOperator(J)

        ra = RandALO(loss=self.loss, jac=jac, y=self.y, y_hat=self.y_hat, rng=self.rng)
        risk = ra.evaluate(self.risk_fun, n_matvecs=200_000)
        self.assertFalse(torch.allclose(ra._normalized_diag_jac_estims, diag[:, None]))

        risk_bks = ra.evaluate_bks(self.risk_fun, n_matvecs=200_000)
        risk_alo = ra.evaluate_alo(self.risk_fun)
        self.assertAlmostEqual(risk, risk_bks, places=3)
        self.assertAlmostEqual(risk, risk_alo, places=3)
        self.assertAlmostEqual(
            risk_alo,
            torch.mean((self.y - self.y_hat) ** 2 / (1 - diag) ** 2).item(),
        )

    def test_logistic(self):
        loss = ml.LogisticLoss()
        X = torch.rand(self.n, self.n, generator=self.rng)
        y = torch.randint(0, 2, (self.n,), generator=self.rng).float() * 2 - 1
        y_hat = torch.randn(self.n, generator=self.rng)

        ds = utils.compute_derivatives(loss, y, y_hat)
        dloss_dy_hat = ds.dloss_dy_hat
        d2loss_dboth = ds.d2loss_dboth
        d2loss_dy_hat2 = ds.d2loss_dy_hat2

        jac = -lo.MatrixOperator(
            X
            @ torch.linalg.solve(
                X.T @ (d2loss_dy_hat2[:, None] * X) + torch.eye(self.n),
                X.T * d2loss_dboth[None, :],
            )
        )
        jac_tilde = X @ torch.linalg.solve(
            X.T @ (d2loss_dy_hat2[:, None] * X) + torch.eye(self.n),
            X.T * d2loss_dy_hat2[None, :],
        )

        ra = RandALO(loss=loss, jac=jac, y=y, y_hat=y_hat, rng=self.rng)

        def risk_fun(y, z):
            return torch.mean(torch.lt(y * z, 0).float()).item()

        jac_tilde_diag = torch.diag(jac_tilde)
        risk = ra.evaluate_alo(risk_fun)
        risk_bks = ra.evaluate_bks(risk_fun, n_matvecs=200)
        self.assertAlmostEqual(risk, risk_bks)

        y_tilde_exact = y_hat + dloss_dy_hat / d2loss_dy_hat2 * jac_tilde_diag / (
            1 - jac_tilde_diag
        )
        self.assertTrue(torch.allclose(ra._y_tilde_exact, y_tilde_exact))
        self.assertAlmostEqual(risk, risk_fun(y, y_tilde_exact))

    def test_more_matvecs(self):
        ra = RandALO(
            loss=self.loss, jac=self.jac, y=self.y, y_hat=self.y_hat, rng=self.rng
        )
        ra.evaluate(self.risk_fun, n_matvecs=3)
        self.assertEqual(ra._n_matvecs, 3)
        self.assertEqual(ra._normalized_diag_jac_estims.shape, (self.n, 3))

        ra.evaluate_bks(self.risk_fun, n_matvecs=5)
        self.assertEqual(ra._n_matvecs, 5)
        self.assertEqual(ra._normalized_diag_jac_estims.shape, (self.n, 5))

        # test that the last matevecs are not used if we request fewer than we have
        ra._normalized_diag_jac_estims[:, -1] = torch.nan
        risk1 = ra.evaluate(self.risk_fun, n_matvecs=2)
        self.assertEqual(ra._n_matvecs, 5)
        self.assertEqual(ra._normalized_diag_jac_estims.shape, (self.n, 5))
        self.assertFalse(np.isnan(risk1))

        risk2 = ra.evaluate_bks(self.risk_fun, n_matvecs=4)
        self.assertEqual(ra._n_matvecs, 5)
        self.assertEqual(ra._normalized_diag_jac_estims.shape, (self.n, 5))
        self.assertFalse(np.isnan(risk2))

        # but if we do use them, we should get nan
        risk3 = ra.evaluate_bks(self.risk_fun, n_matvecs=6)
        self.assertEqual(ra._n_matvecs, 6)
        self.assertEqual(ra._normalized_diag_jac_estims.shape, (self.n, 6))
        self.assertTrue(np.isnan(risk3))

    def test_uniform_map_estimates(self):
        ra = RandALO(
            loss=self.loss, jac=self.jac, y=self.y, y_hat=self.y_hat, rng=self.rng
        )
        ra.evaluate(self.risk_fun, n_matvecs=3)

        mus = ra._normalized_diag_jac_estims.mean(dim=1, keepdim=True)
        stds = ra._normalized_diag_jac_stds[:, None]
        ms = torch.arange(1, 4, dtype=torch.float32)[None, :]
        sigmas = stds / torch.sqrt(ms)
        uniform_map_estimates = ra._uniform_map_estimates(mus, ms)
        truncnorm_means = truncnorm.truncnorm_mean(
            mus, sigmas, torch.tensor([[0.0]]), torch.tensor([[1.0]])
        )
        self.assertTrue(torch.allclose(uniform_map_estimates, truncnorm_means))

        # add another dimension
        mus = mus[..., None]
        stds = stds[..., None]
        ms = ms[..., None]
        sigmas = stds / torch.sqrt(ms)
        uniform_map_estimates = ra._uniform_map_estimates(mus, ms)
        truncnorm_means = truncnorm.truncnorm_mean(
            mus, sigmas, torch.tensor([[[0.0]]]), torch.tensor([[[1.0]]])
        )
        self.assertTrue(torch.allclose(uniform_map_estimates, truncnorm_means))


if __name__ == "__main__":
    unittest.main()



import unittest

import numpy as np
import scipy
import scipy.special
import sklearn.linear_model
import sklearn.linear_model._coordinate_descent
import torch

from randalo import RandALO
from randalo import modeling_layer as ml
from randalo import utils


class TestSklearnRandALO(unittest.TestCase):
    def setUp(self):
        np.random.seed(0)
        self.n = 10
        self.p = 8
        self.rng = np.random.default_rng(0)
        self.X = self.rng.normal(0, 1, (self.n, self.p))
        self.beta = self.rng.normal(0, 1 / np.sqrt(self.p), (self.p))
        self.y = self.X @ self.beta + self.rng.normal(0, 1, (self.n))
        self.y_bin = (
            self.rng.uniform(0, 1, (self.n,))
            < scipy.special.expit(4 * self.X @ self.beta)
        ).astype(int)

    def compute_dy_hat(self, model, dy):
        model.fit(self.X, self.y)
        y_hat = model.predict(self.X)
        model.fit(self.X, self.y + dy)
        y_hat2 = model.predict(self.X)
        return y_hat2 - y_hat

    def assertRandomJacobianDirectionAlmostEqual(
        self, model, jac, epsilon=1e-6, sim_thresh=0.99, norm_rtol=1e-3
    ):
        dy = self.rng.normal(0, epsilon, (self.n,))
        dy_hat = utils.to_tensor(self.compute_dy_hat(model, dy))
        dy = utils.to_tensor(dy)

        dy_hat = dy_hat / torch.norm(dy)
        djac = jac @ dy / torch.norm(dy)
        self.assertTrue(
            torch.allclose(torch.norm(dy_hat), torch.norm(djac), rtol=norm_rtol)
        )

        if torch.norm(dy_hat) == 0 or torch.norm(djac) == 0:
            return

        sim = dy_hat / torch.norm(dy_hat) @ djac / torch.norm(djac)
        self.assertGreaterEqual(sim, sim_thresh)

    def get_randalo_jac(self, model, y=None):
        if y is None:
            y = self.y
        ra = RandALO.from_sklearn(model, self.X, y)
        return ra._jac @ torch.eye(self.n)

    def test_linear_regression(self):
        lr = sklearn.linear_model.LinearRegression(fit_intercept=False)
        lr.fit(self.X, self.y)

        # check against theoretical ground truth
        Q, _ = np.linalg.qr(self.X)
        jac = utils.to_tensor(Q @ Q.T)
        ra_jac = self.get_randalo_jac(lr)
        self.assertTrue(torch.allclose(jac, ra_jac, atol=1e-6))

        # check numerically
        self.assertRandomJacobianDirectionAlmostEqual(lr, ra_jac)

    def test_ridge(self):
        alphas = np.logspace(-1, 1, 10)

        for alpha in alphas:
            ridge = sklearn.linear_model.Ridge(alpha=alpha, fit_intercept=False)
            ridge.fit(self.X, self.y)

            # check against theoretical ground truth
            jac = utils.to_tensor(
                self.X
                @ np.linalg.solve(
                    self.X.T @ self.X + alpha * np.eye(self.p),
                    self.X.T,
                )
            )
            ra_jac = self.get_randalo_jac(ridge)
            self.assertTrue(torch.allclose(jac, ra_jac, atol=1e-6))

            # check against the wrong Jacobian
            bad_jac = utils.to_tensor(
                self.X
                @ np.linalg.solve(
                    self.X.T @ self.X + 2 * alpha * np.eye(self.p),
                    self.X.T,
                )
            )
            self.assertFalse(torch.allclose(bad_jac, ra_jac, atol=1e-6))

            # check numerically
            self.assertRandomJacobianDirectionAlmostEqual(ridge, ra_jac)

    def test_lasso(self):
        # get a path of parameters and shrink a bit so we aren't at the max
        alphas = (
            sklearn.linear_model._coordinate_descent._alpha_grid(
                self.X,
                self.y,
                fit_intercept=False,
                l1_ratio=1.0,
                n_alphas=10,
            )
            * 0.99
        )

        # store the unique numbers of nonzeros over the alphas
        nnzs = set()

        # fit a lasso model for each alpha and check Jacobian
        for alpha in alphas:
            lasso = sklearn.linear_model.Lasso(
                alpha=alpha, fit_intercept=False, tol=1e-8
            )
            lasso.fit(self.X, self.y)
            lasso_lars = sklearn.linear_model.LassoLars(
                alpha=alpha, fit_intercept=False, max_iter=10000
            )
            lasso_lars.fit(self.X, self.y)

            mask = lasso.coef_ != 0
            X_mask = self.X[:, mask]
            nnzs.add(np.sum(mask))

            # check against theoretical ground truth
            Q, _ = np.linalg.qr(X_mask)
            jac = utils.to_tensor(Q @ Q.T)
            ra_jac = self.get_randalo_jac(lasso)
            ra_jac_lars = self.get_randalo_jac(lasso_lars)
            self.assertTrue(torch.allclose(jac, ra_jac, atol=1e-6))
            self.assertTrue(torch.allclose(jac, ra_jac_lars, atol=1e-6))

            # check numerically
            self.assertRandomJacobianDirectionAlmostEqual(
                lasso, ra_jac, sim_thresh=0.999, norm_rtol=1e-3
            )
            self.assertRandomJacobianDirectionAlmostEqual(
                lasso_lars, ra_jac, sim_thresh=0.9999, norm_rtol=1e-2
            )

        # make sure we've actually checked different sparsity patterns
        self.assertTrue(len(nnzs) > 3)

    def test_elastic_net(self):
        # get a path of parameters and shrink a bit
        l1_ratio = 0.5
        alphas = (
            sklearn.linear_model._coordinate_descent._alpha_grid(
                self.X,
                self.y,
                fit_intercept=False,
                l1_ratio=l1_ratio,
                n_alphas=10,
            )
            * 0.99
        )

        # store the unique numbers of nonzeros over the alphas
        nnzs = set()

        # fit a lasso model for each alpha and check Jacobian
        for alpha in alphas:
            enet = sklearn.linear_model.ElasticNet(
                alpha=alpha, l1_ratio=l1_ratio, fit_intercept=False, tol=1e-8
            )
            enet.fit(self.X, self.y)

            mask = enet.coef_ != 0
            X_mask = self.X[:, mask]
            n_mask = np.sum(mask)
            nnzs.add(n_mask)

            # check against theoretical ground truth
            jac = utils.to_tensor(
                X_mask
                @ np.linalg.solve(
                    X_mask.T @ X_mask / self.n
                    + alpha * (1 - l1_ratio) * np.eye(n_mask),
                    X_mask.T / self.n,
                )
            )
            ra_jac = self.get_randalo_jac(enet)
            self.assertTrue(torch.allclose(jac, ra_jac, atol=1e-6))

            # check numerically
            self.assertRandomJacobianDirectionAlmostEqual(enet, ra_jac, norm_rtol=0.02)

        # make sure we've actually checked different sparsity patterns
        self.assertTrue(len(nnzs) > 3)

    def test_logistic(self):
        param_dicts = [
            {"penalty": None, "eye_scale": 0.0},
            {"penalty": "l1", "C": 1.0, "eye_scale": 0.0},
            {"penalty": "l2", "C": 1.0, "eye_scale": 1.0},
            {"penalty": "elasticnet", "C": 1.0, "l1_ratio": 0.5, "eye_scale": 0.5},
        ]
        lr = sklearn.linear_model.LogisticRegression(
            tol=1e-5, solver="saga", max_iter=100000, fit_intercept=False
        )

        y = utils.to_tensor(self.y_bin * 2 - 1)
        y_labels = np.array(["a", "b"])[self.y_bin]
        nnzs = set()

        for param_dict in param_dicts:
            eye_scale = param_dict.pop("eye_scale")
            lr.set_params(**param_dict)
            lr.fit(self.X, y_labels)

            mask = lr.coef_[0, :] != 0
            X_mask = self.X[:, mask]
            n_mask = np.sum(mask)
            nnzs.add(n_mask)

            y_hat = utils.to_tensor(lr.decision_function(self.X))
            ds = utils.compute_derivatives(ml.LogisticLoss(), y, y_hat)
            d2loss_dboth = ds.d2loss_dboth.numpy()
            d2loss_dy_hat2 = ds.d2loss_dy_hat2.numpy()

            jac = utils.to_tensor(
                -X_mask
                @ np.linalg.solve(
                    X_mask.T @ (d2loss_dy_hat2[:, None] * X_mask)
                    + eye_scale / self.n * np.eye(n_mask),
                    X_mask.T * d2loss_dboth[None, :],
                )
            )
            ra_jac = self.get_randalo_jac(lr, y=y_labels)
            atol = 1e-2 if param_dict["penalty"] is None else 1e-6
            self.assertTrue(torch.allclose(jac, ra_jac, atol=atol))

            # check that wrong y = wrong Jacobian
            ds_bad = utils.compute_derivatives(ml.LogisticLoss(), -y, y_hat)
            d2loss_dboth_bad = ds_bad.d2loss_dboth.numpy()
            d2loss_dy_hat2_bad = ds_bad.d2loss_dy_hat2.numpy()

            jac_bad = utils.to_tensor(
                -X_mask
                @ np.linalg.solve(
                    X_mask.T @ (d2loss_dy_hat2_bad[:, None] * X_mask)
                    + eye_scale / self.n * np.eye(n_mask),
                    X_mask.T * d2loss_dboth_bad[None, :],
                )
            )
            self.assertFalse(torch.allclose(jac_bad, ra_jac, atol=atol))


if __name__ == "__main__":
    unittest.main()



import unittest

import linops as lo
import cvxpy as cp
import numpy as np
import torch
from cvxpylayers.torch import CvxpyLayer

from randalo import modeling_layer as ml, reductions


class TestReductions(unittest.TestCase):
    def setUp(self):
        self.n = 10
        self.rng = np.random.default_rng(0x219A)
        self.loss = ml.MSELoss()
        self.regularizer = 0.01 * (0.5 * ml.SquareRegularizer() + ml.L1Regularizer())
        self.X = self.rng.standard_normal((self.n, 3 * self.n))
        self.y = self.X[:, 0] + 0.01 * self.rng.standard_normal(self.n)

    def test_transform_to_cvxpy_and_test_jacobian(self):
        # TODO fix segemntation fault?
        return

        b = cp.Variable(3 * self.n)
        y = cp.Parameter(self.n)
        prob = reductions.transform_model_to_cvxpy(
            self.loss, self.regularizer, self.X, y, b
        )
        self.layer = CvxpyLayer(prob, parameters=[y], variables=[b])

        y_torch = torch.tensor(self.y, requires_grad=True)
        J = reductions.Jacobian(
            self.y, self.X, lambda: b.value, self.loss, self.regularizer
        )

        y.value = self.y
        prob.solve()
        y_hat_torch = torch.from_numpy(self.X) @ self.layer(y_torch)[0]

        z = self.rng.standard_normal(self.n)
        Jz = J @ z
        (y_hat_torch @ torch.from_numpy(z)).backward()
        assert torch.allclose(Jz, y_torch.grad.to(torch.float32), atol=1e-4, rtol=1e-2)


if __name__ == "__main__":
    unittest.main()



import numpy as np
import torch

import sklearn.base
import sklearn.linear_model
import sklearn.utils.validation

from . import modeling_layer as ml
from . import reductions


def map_sklearn(
    model: sklearn.base.BaseEstimator = None,
    X: torch.Tensor | np.ndarray = None,
    y: torch.Tensor | np.ndarray | list = None,
) -> tuple[ml.Loss, reductions.Jacobian, torch.Tensor, torch.Tensor]:
    sklearn.utils.validation.check_is_fitted(model)
    n = X.shape[0]

    def solution_func():
        return model.coef_

    match model:
        case sklearn.linear_model.LinearRegression():
            loss = ml.MSELoss()
            # TODO create trivial regularizer
            reg = 0 * ml.SquareRegularizer()
            y_hat = model.predict(X)

        case sklearn.linear_model.Ridge():
            loss = ml.MSELoss()
            reg = model.alpha / n * ml.SquareRegularizer()
            y_hat = model.predict(X)

        case sklearn.linear_model.Lasso() | sklearn.linear_model.LassoLars():
            loss = ml.MSELoss()
            reg = 2.0 * model.alpha * ml.L1Regularizer()
            y_hat = model.predict(X)

        case sklearn.linear_model.ElasticNet():
            loss = ml.MSELoss()
            reg = model.alpha * (
                2.0 * model.l1_ratio * ml.L1Regularizer()
                + (1 - model.l1_ratio) * ml.SquareRegularizer()
            )
            y_hat = model.predict(X)

        case sklearn.linear_model.LogisticRegression():
            # TODO: implement sample weight
            if len(model.classes_) > 2:
                raise ValueError("Only binary classification is supported.")

            loss = ml.LogisticLoss()
            y = ((model.classes_[None, :] == y[:, None]) * np.array([[-1, 1]])).sum(1)
            match model.penalty:
                case None:
                    # TODO: create trivial regularizer
                    reg = 0 * ml.SquareRegularizer()
                case "l1":
                    reg = ml.L1Regularizer()
                case "l2":
                    reg = 0.5 * ml.SquareRegularizer()
                case "elasticnet":
                    reg = (
                        model.l1_ratio * ml.L1Regularizer()
                        + 0.5 * (1 - model.l1_ratio) * ml.SquareRegularizer()
                    )
            reg = reg * 1 / model.C / n
            y_hat = model.decision_function(X)

            def solution_func():
                return model.coef_[0, :]

    jac = reductions.Jacobian(y, X, solution_func, loss, reg, inverse_method=None)

    return loss, jac, y, y_hat



from typing import Callable

import linops as lo
import numpy as np
import sklearn.base
import torch

from . import modeling_layer as ml
from . import sklearn_integration as ski
from . import truncnorm
from . import utils


class RandALO(object):
    def __init__(
        self,
        loss: ml.Loss = None,
        jac: lo.LinearOperator = None,
        y: torch.Tensor | np.ndarray = None,
        y_hat: torch.Tensor | np.ndarray = None,
        dtype: torch.dtype = torch.float32,
        device: torch.device = None,
        rng: torch.Generator = None,
    ):
        """Initialize the RandALO object.

        Parameters
        ----------
        loss : Loss
            Loss function.
        jac : LinearOperator
            Jacobian operator.
        y : torch.Tensor | np.ndarray
            True labels.
        y_hat : torch.Tensor | np.ndarray
            Predicted values.
        dtype : torch.dtype, optional
            Data type for tensors, by default torch.float32.
        device : torch.device, optional
            Device for tensors, by default None.
        rng : torch.Generator, optional
            Random number generator, by default None.
        """

        if loss is None:
            raise ValueError("loss function must be provided")
        self._loss = loss

        if jac is None:
            raise ValueError("Jacobian operator must be provided")
        self._jac = jac

        if y is None:
            raise ValueError("label values must be provided")
        y = utils.to_tensor(y, dtype=dtype, device=device)

        if y_hat is None:
            raise ValueError("predicted values must be provided")
        y_hat = utils.to_tensor(y_hat, dtype=dtype, device=device)

        self._dtype = dtype
        self._device = y.device
        if rng is None:
            rng = torch.Generator(device=device)
            rng.seed()
        self._rng = rng

        # check dtypes and devices
        # assert self._jac.dtype == self._dtype
        # assert self._jac.device == self._device

        # compute derivatives of loss function
        (
            self._y,
            self._y_hat,
            self._dloss_dy_hat,
            self._d2loss_dboth,
            self._d2loss_dy_hat2,
        ) = utils.compute_derivatives(self._loss, y, y_hat)

        # precompute some quantities for working with generic form
        # a = dloss_dy_hat
        # b = d2loss_dy_hat2
        # c = d2loss_dboth
        self._a_c = self._dloss_dy_hat / self._d2loss_dy_hat2
        self._c_b = self._d2loss_dy_hat2 / self._d2loss_dboth

        # the normalized Jacobian has diagonal values in the range [0, 1]
        self._normalized_jac = self._jac @ lo.DiagonalOperator(-self._c_b)

        self._n_matvecs = 0
        self._normalized_diag_jac_estims = None
        self._y_tilde_exact = None

    def evaluate(
        self,
        risk_fun: Callable[[torch.Tensor, torch.Tensor], float],
        n_matvecs: int = 100,
        subsets: list[list[int]] | int = 50,
    ) -> float:
        """Evaluate the risk function using RandALO:

            R = 1 / n * sum_i risk(y_i, y_hat_i).

        Parameters
        ----------
        risk_fun : Callable[[torch.Tensor, torch.Tensor], torch.Tensor]
            The risk function to evaluate.
        n_matvecs : int, optional
            The number of Jacobian–vector products to compute for the RandALO
            method, by default 100.
        subsets : list[list[int]] | int, optional
            The subsets of Jacobian–vector products to use for the debiasing
            step in RandALO. If an integer is provided, the subsets are chosen
            randomly with sizes uniformly taken between `n_matvecs // 2` and
            `n_matvecs`. By default 50.

        Returns
        -------
        float
            The risk estimate.
        """
        self._do_diag_jac_estims_upto(n_matvecs)

        # compute BKS estimates for subsets of Jacobian–vector products
        if isinstance(subsets, int):
            subsets = [
                torch.randperm(n_matvecs, generator=self._rng)[:m]
                for m in torch.linspace(n_matvecs // 2, n_matvecs, subsets, dtype=int)
            ]
        mixing_matrix = utils.create_mixing_matrix(n_matvecs, subsets)
        mus = self._normalized_diag_jac_estims[:, :n_matvecs] @ mixing_matrix
        m_primes = torch.sum(mixing_matrix > 0, dim=0, keepdim=True)
        normalized_diag_jacs = self._uniform_map_estimates(mus, m_primes)
        y_tildes = [
            self._y_tilde_from_normalized_jac(normalized_diag_jacs[:, j])
            for j in range(len(subsets))
        ]
        risks = [risk_fun(self._y, y_tilde) for y_tilde in y_tildes]
        # return utils.robust_y_intercept(1 / m_primes, risks), m_primes, risks
        return utils.robust_y_intercept(1 / m_primes, risks)

    def evaluate_bks(
        self,
        risk_fun: Callable[[torch.Tensor, torch.Tensor], torch.Tensor],
        n_matvecs: int = 100,
    ) -> float:
        """Evaluate the risk function using the plug-in BKS estimate.

        Parameters
        ----------
        risk_fun : Callable[[torch.Tensor, torch.Tensor], float]
            The risk function to evaluate.
        n_matvecs : int, optional
            The number of Jacobian–vector products to compute for the BKS
            method, by default 100.

        Returns
        -------
        float
            The risk estimate.
        """

        self._do_diag_jac_estims_upto(n_matvecs)

        mus = self._normalized_diag_jac_estims[:, :n_matvecs].mean(dim=1)
        normalized_diag_jac_bks = self._uniform_map_estimates(
            mus, utils.unsqueeze_scalar_like(n_matvecs, mus)
        )

        y_tilde = self._y_tilde_from_normalized_jac(normalized_diag_jac_bks)
        return risk_fun(self._y, y_tilde)

    def evaluate_alo(
        self, risk_fun: Callable[[torch.Tensor, torch.Tensor], torch.Tensor]
    ) -> float:
        """Perform exact ALO.

        Parameters
        ----------
        risk_fun : Callable[[torch.Tensor, torch.Tensor], float]
            The risk function to evaluate.

        Returns
        -------
        float
            The risk estimate.

        """
        if self._y_tilde_exact is None:
            if hasattr(self._jac, "diag"):
                self._y_tilde_exact = self._y_tilde_from_jac(self._jac.diag())
            else:
                self._y_tilde_exact = self._y_tilde_from_normalized_jac(
                    torch.diag(self._normalized_jac @ torch.eye(self._y.shape[0]))
                )

        return risk_fun(self._y, self._y_tilde_exact)

    def _y_tilde_from_normalized_jac(
        self, normalized_jac: torch.Tensor
    ) -> torch.Tensor:
        """Compute the corrected predictions using the normalized Jacobian.

        Parameters
        ----------
        normalized_jac : torch.Tensor
            The transformed diagonal of the Jacobian.

        Returns
        -------
        torch.Tensor
            The corrected predictions.
        """
        normalized_jac = torch.clamp(normalized_jac, 0, 1)
        return self._y_hat + self._a_c * normalized_jac / (1 - normalized_jac)

    def _y_tilde_from_jac(self, diag_jac: torch.Tensor) -> torch.Tensor:
        """Convenience method for computing the corrected predictions.

        First transforms the diagonal of the Jacobian into a generic form,
        then computes the corrected predictions.

        Parameters
        ----------
        diag_jac : Tensor
            The diagonal of the Jacobian.

        Returns
        -------
        Tensor
            The corrected predictions.
        """

        return self._y_tilde_from_normalized_jac(-diag_jac * self._c_b)

    def _get_matvecs(self, n_matvecs: int) -> tuple[torch.Tensor, torch.Tensor]:
        """Compute random Jacobian–vector products using random Rademacher
        vectors.

        Parameters
        ----------
        n_matvecs : int
            The number of samples to compute.

        Returns
        -------
        [Tensor, Tensor]
            The matrix–vector products and the random vectors.
        """
        Omega = (
            torch.randint(
                0,
                2,
                (self._y.shape[0], n_matvecs),
                generator=self._rng,
                dtype=self._dtype,
            )
            * 2.0
            - 1
        )
        return self._normalized_jac @ Omega, Omega

    def _do_diag_jac_estims_upto(self, n_matvecs: int) -> None:
        """Compute more diagonal Jacobian estimates.

        Parameters
        ----------
        n_matvecs : int
            The desired total number of samples.
        """

        if n_matvecs <= self._n_matvecs:
            return

        # compute the matrix–vector products
        matvecs, Omega = self._get_matvecs(n_matvecs - self._n_matvecs)

        # update the diagonal Jacobian estimates
        normalized_diag_jac_estims = matvecs * Omega
        if self._normalized_diag_jac_estims is None:
            self._normalized_diag_jac_estims = normalized_diag_jac_estims
        else:
            self._normalized_diag_jac_estims = torch.cat(
                (self._normalized_diag_jac_estims, normalized_diag_jac_estims), dim=1
            )
        self._n_matvecs = n_matvecs

        # compute statistics for truncated normal MAP estimation
        self._normalized_diag_jac_stds = self._normalized_diag_jac_estims.std(dim=1)

    def _uniform_map_estimates(self, mus: torch.Tensor, ms: torch.Tensor):
        """Compute MAP estimates with a Uniform[0, 1] prior given sample means.
        Uses the standard deviations computed from all available samples, even
        for sample means computed from a subset of samples.

        Parameters
        ----------
        mus : torch.Tensor
            The sample means of the diagonal estimates.
        ms: torch.Tensor
            The number of samples, broadcastable with `mus`.

        Returns
        -------
        torch.Tensor
            The MAP estimates.
        """
        return truncnorm.truncnorm_mean(
            mus,
            self._normalized_diag_jac_stds.reshape(-1, *([1] * (mus.ndim - 1)))
            / torch.sqrt(ms),
            utils.unsqueeze_scalar_like(0.0, mus),
            utils.unsqueeze_scalar_like(1.0, mus),
        )

    @classmethod
    def from_sklearn(
        cls,
        model: sklearn.base.BaseEstimator = None,
        X: torch.Tensor | np.ndarray = None,
        y: torch.Tensor | np.ndarray = None,
    ) -> "RandALO":
        """Instantiate a RandALO object from a scikit-learn model.

        Automatically converts a fitted scikit-learn model to a RandALO object
        by mapping the model and parameters to the appropriate loss and Jacobian.

        Parameters
        ----------
        model : sklearn.base.BaseEstimator
            The fitted scikit-learn estimator.
        X : torch.Tensor | np.ndarray
            The training data that the model was fitted on.
        y : torch.Tensor | np.ndarray | list
            The training labels that the model was fitted on.

        Returns
        -------
        RandALO
            The instantiated RandALO object.
        """
        loss, jac, y, y_hat = ski.map_sklearn(model, X, y)
        return cls(loss=loss, jac=jac, y=y, y_hat=y_hat)



import functools
from typing import Callable, Literal
from dataclasses import dataclass, field

import numpy as np
import linops as lo
import torch

from . import modeling_layer as ml
from . import utils

def gen_cvxpy_jacobian(loss, regularizer, X, variable, y, inversion_method=None):
    prob = transform_model_to_cvxpy(loss, regularizer, X, y, variable)
    J = Jacobian(y, X, lambda: variable.value, loss, regularizer, inversion_method=None)
    return prob, J

def transform_model_to_cvxpy(loss, regularizer, X, y, variable):
    import cvxpy as cp
    return cp.Problem(
        cp.Minimize(
            loss.to_cvxpy(y, X @ regularizer) +
            regularizer.to_cvxpy(variable)
        )
    )


class Jacobian(lo.LinearOperator):
    solution_func: Callable[[], torch.Tensor]
    loss: ml.Loss
    regularizer: ml.Sum | ml.Regularizer
    inverse_method: Literal[None, "minres", "cholesky"]

    supports_operator_matrix = True

    def __init__(self, y, X, solution_func, loss, regularizer, inverse_method=None):
        self.solution_func = solution_func
        self.loss = loss
        self.regularizer = regularizer
        self.inverse_method = inverse_method
        self.y = utils.to_tensor(y)
        self.X = utils.to_tensor(X)

    @property
    def _shape(self):
        n = self.y.shape[0]
        return (n, n)

    _diag: torch.Tensor = None

    # @functools.cached_property
    def diag(self):
        return torch.diag(self @ torch.eye(self.shape[1]))

    def _matmul_impl(self, rhs):
        rhs = utils.to_tensor(rhs)
        needs_squeeze = False
        if len(rhs.shape) == 1:
            rhs = rhs.unsqueeze(-1)
            needs_squeeze = True
        beta_hat = utils.to_tensor(self.solution_func())
        y = self.y
        X = self.X
        _, _, _, d2loss_dboth, d2loss_dy_hat2 = utils.compute_derivatives(
            self.loss, y, X @ beta_hat
        )

        constraints, hessians, mask = \
                self.regularizer.get_constraint_hessian_mask(beta_hat)
        if mask is not None:
            X_mask = X[:, mask]
        else:
            X_mask = X
        rhs_scaled = -d2loss_dboth[:, None] * rhs

        if constraints is None and hessians is None:
            sqrt_d2loss_dy_hat2 = torch.sqrt(d2loss_dy_hat2)[:, None]
            tilde_X = sqrt_d2loss_dy_hat2 * X_mask
            Q, _ = torch.linalg.qr(tilde_X)
            return (
                Q @ (Q.T @ (rhs_scaled / sqrt_d2loss_dy_hat2))
            ) / sqrt_d2loss_dy_hat2
        elif constraints is None:
            # TODO: double check this doesn't need additional scaling
            kkt_rhs = X_mask.T @ rhs_scaled
            if mask is not None:
                hessians_mask = hessians[mask, :][:, mask]
            else:
                hessians_mask = hessians
            P = X_mask.T @ (d2loss_dy_hat2[:, None] * X_mask) + hessians_mask
            R = torch.linalg.cholesky(P, upper=True)
            v = torch.linalg.solve_triangular(
                R, torch.linalg.solve_triangular(R.T, kkt_rhs, upper=False), upper=True
            )
        else:
            # TODO: double check this doesn't need additional scaling
            if mask is not None:
                constraints_mask = constraints[:, mask]
            else:
                constraints_mask = constraints
            n, m = constraints_mask.shape
            if n >= m:
                _, N = torch.linalg.qr(constraints_mask, mode="r")
            else:
                N = constraints_mask

            if hessians is None:
                tilde_X = torch.sqrt(d2loss_dy_hat2)[:, None] * X_mask
                _, P_R = torch.linalg.qr(tilde_X, mode="r")
            else:
                if mask is not None:
                    hessians_mask = hessians[mask, :][:, mask]
                else:
                    hessian_mask = hessians
                P = X_mask.T @ (d2loss_dy_hat2[:, None] * X_mask) + hessians_mask
                P_R = torch.linalg.cholesky(P, upper=True)

            S = self.D_nmask @ torch.linalg.solve_triangular(
                P_R,
                torch.linalg.solve_triangular(P_R.T, kkt_rhs, upper=False),
                upper=True,
            )
            S_R = torch.linalg.cholesky(S, upper=True)
            NPinvRhs = N @ torch.linalg.solve_triangular(
                P_R,
                torch.linalg.solve_triangular(P_R.T, kkt_rhs, upper=False),
                upper=True,
            )
            nu = torch.linalg.solve_triangular(
                S_R,
                torch.linalg.solve_triangular(S_R.T, -NPinvRhs, upper=False),
                upper=True,
            )
            v = torch.linalg.solve_triangular(
                P_R,
                torch.linalg.solve_triangular(P_R.T, kkt_rhs + N.T @ nu, upper=False),
                upper=True,
            )
        out = X_mask @ v
        return out if not needs_squeeze else out.squeeze(-1)



from abc import ABC, abstractmethod
from dataclasses import dataclass, field

import cvxpy as cp
import numpy as np
import torch


@dataclass
class HyperParameter:
    parameter: cp.Parameter = field(default_factory=cp.Parameter)
    scale: float = field(init=False, default=1.0)

    def __mul__(self, r):
        # TODO: __mul__ should not have side-effects
        if isinstance(r, float | np.float32):
            return HyperParameter(parameter=self.parameter, scale=self.scale * r)
        else:
            return NotImplemented

    @property
    def value(self):
        return self.scale * self.parameter.value

    @value.setter
    def value(self, val):
        assert self.scale == 1.0, "Cannot set the value of a scaled parameter"
        self.parameter.value = val


@dataclass
class Regularizer(ABC):
    linear: np.ndarray | list[int] = field(default=None)
    scale: float = field(init=False, default=1.0)
    parameter: HyperParameter = field(init=False, default=None)

    def __mul__(self, r):
        if isinstance(r, HyperParameter):
            if self.parameter is not None:
                raise TypeError("Cannot have multiple parameters")
            out = type(self)(linear=self.linear)
            out.scale = self.scale
            out.parameter= r
            return out
        # elif isinstance(r, float | np.float32): # <- didn't allow integers or other floats...
        else:
            out = type(self)(linear=self.linear)
            out.scale = self.scale * r
            out.parameter = self.parameter
            return out
        # else:
        #     raise TypeError("Multiply must be with either a scalar or HyperParameter")
        return NotImplemented

    def __rmul__(self, r):
        return self * r

    def __truediv__(self, r):
        return self * (1 / r)

    def __add__(self, rhs):
        if rhs == 0:
            return self
        if isinstance(rhs, Sum):
            return Sum([self] + rhs.exprs)
        elif isinstance(rhs, Regularizer):
            return Sum([self, rhs])
        else:
            raise TypeError("Addend must be a Regularizer or Sum")

    def __radd__(self, lhs):
        return self + lhs

    @abstractmethod
    def to_cvxpy(self, variable, func):
        expr = func(obj.linear @ variable if obj.linear is not None else variable)
        if obj.parameter is None:
            return obj.scale * expr
        else:
            return obj.scale * obj.parameter.scale * obj.parameter.parameter * expr

    @abstractmethod
    def get_constraint_hessian_mask(self, beta_hat, epsilon=1e-6):
        """
        Returns two matrices and a vector[bool] (or None if they would be the
        0 matrix or all true) such that
        the hessian of this regularizer can be viewed as 
            infinity * (first matrix) + (second matrix) + infty * (third matrix)
        first matrix is called the "constraint matrix" second matrix is called
        the "hessian matrix"

        The third matrix is given by
            diag = torch.zero(n)
            diag[mask] = 1
            third matrix = torch.diag(diag)
        ie any entry associated with a False in mask is held to always be zeros
        """

    def _scale(self):
        if self.parameter is not None:
            return self.scale * self.parameter.value
        else:
            return self.scale


class SquareRegularizer(Regularizer):
    def to_cvxpy(self, variable):
        super().to_cvxpy(variable, cp.sum_squares)

    def get_constraint_hessian_mask(self, beta_hat, epsilon=1e-6):
        scale = self._scale()
        if scale == 0.0:
            return None, None, None
        if self.linear is None:
            return None, torch.diag(
                    2 * scale * torch.ones_like(beta_hat, dtype=beta_hat.dtype)), None
        elif isinstance(linear, list):
            diag = torch.zeros_like(beta_hat)
            diag[linear] = scale
            return None, torch.diag(diag), None
        else:
            A = utils.to_tensor(linear)
            return None, torch.diag(scale * (A.mT @ A)), None


class L1Regularizer(Regularizer):
    def to_cvxpy(self, variable):
        super().to_cvxpy(variable, cp.norm1)

    def get_constraint_hessian_mask(self, beta_hat, epsilon=1e-6):
        scale = self._scale()
        mask = torch.ones_like(beta_hat, dtype=bool)
        if self.linear is None:
            mask[torch.abs(beta_hat) <= epsilon] = False
            return None, None, mask
        elif isinstance(linear, list):
            mask[linear][torch.abs(beta_hat[linear]) <= epsilon] = False
            return None, None, mask
        else:
            A = utils.from_numpy(linear)
            return A[torch.abs(A @ beta_hat) <= epsilon, :], None, None


class L2Regularizer(Regularizer):
    def to_cvxpy(self, variable):
        super().to_cvxpy(variable, cp.norm2)

    def get_constraint_hessian_mask(self, beta_hat, epsilon=1e-6):
        linear = self.linear
        if self.linear is None:
            norm = torch.linalg.norm(beta_hat)
            if norm <= epsilon:
                mask = torch.zeros_like(beta_hat, dtype=bool)
                return None, None, mask

            tilde_beta_hat_2d = torch.atleast_2d(beta_hat).T / norm
            hessian = self._scale() * (torch.eye(beta_hat.shape) - beta_hat_2d @ beta_hat_2d.T)
            return None, hessian, None
        elif isinstance(linear, list):
            norm = torch.linalg.norm(beta_hat[linear])
            if norm <= epsilon:
                mask = torch.ones_like(beta_hat, dtype=bool)
                mask[linear] = False
                return None, None, mask
            tilde_b = torch.atleast_2d(torch.zeros_like(beta_hat)).T / norm
            tilde_b[linear] = beta_hat[linear]
            diag = torch.zero_like(beta_hat)
            diag[linear] = 1.0
            return None, self._scale() * (torch.diag(diag) - tilde_b @ tilde_b.T), None

        else:
            Lb = linear @ beta_hat
            norm = torch.linalg.norm(Lb)
            tilde_Lb = torch.atleast_2d(Lb).T / norm
            if Lb <= epsilon:
                return linear, None, None
            return None, self._scale() * linear.T @ (
                    torch.eye(Lb.shape[0]) - Lb_2d @ Lb_2d.T) @ linear, None


class HuberRegularizer(Regularizer):
    def to_cvxpy(self, variable):
        super().to_cvxpy(variable, cp.huber)


@dataclass
class Sum:
    exprs: list[Regularizer]

    def __mul__(self, r):
        return Sum([r * expr for expr in self.exprs])

    def __rmul__(self, r):
        return self * r

    def __truediv__(self, r):
        return self * (1 / r)

    def __add__(self, r):
        if isinstance(r, Sum):
            return Sum(r.exprs + self.exprs)
        else:
            return NotImplemented

    def to_cvxpy(self, variable):
        return cp.sum([term.to_cvxpy(variable) for term in terms])

    def get_constraint_hessian_mask(self, beta_hat, epsilon=1e-6):
        constraints = []
        hessians = []
        mask = torch.ones_like(beta_hat, dtype=bool)
        for reg in self.exprs:
            cons, hess, m = reg.get_constraint_hessian_mask(beta_hat, epsilon)
            if cons is not None:
                constraints.append(cons)
            if hess is not None:
                hessians.append(hess)
            if m is not None:
                mask &= m

        constraints = torch.vstack(constraints) if len(constraints) > 0 else None
        hessians = sum(hessians) if len(hessians) > 0 else None
        return constraints, hessians, mask
 

class Loss(ABC):
    def __call__(self, y, z):
        return torch.mean(self.func(y, z))

    @abstractmethod
    def func(self, y, z):
        pass

    @abstractmethod
    def to_cvxpy(self, X, y, variable):
        pass


class LogisticLoss(Loss):
    def func(self, y, z):
        return torch.log(1 + torch.exp(-y * z))

    def to_cvxpy(self, y, z):
        return cp.sum(cp.logistic(-cp.multiply(y, z))) / np.prod(y.shape)
                 

class MSELoss(Loss):
    def func(self, y, z):
        return (y - z) ** 2

    def to_cvxpy(self, y, z):
        return cp.sum_squares(y - z) / np.prod(y.shape)
